
from .abstractdocker import AbstractDockerAwareHandler
from ..result import CommandExecutionResult
from ..entity.definition.sql import PostgreSQLDefinition, PostgreSQLBaseBackupDefinition


class PostgreSQLDumpBackup(AbstractDockerAwareHandler):
    """
<sphinx>
postgres
--------

PostgreSQL backup using traditional dump & import method

On restore:

    - Closes all client connections
    - Blocks access to the database by issuing connection limit = 0
    - After successful or failed restore sets connection limit = -1
      (currently does not support checking previous limit)
    """

    def _get_definition(self) -> PostgreSQLDefinition:
        return self._definition

    def validate_before_creating_backup(self):
        pass

    def receive_backup_stream(self):
        return self.execute_command_in_proper_context(
            command=self._get_definition().get_dump_command() + ' | gzip',
            mode='backup'
        )

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        self.set_connection_limit_on_databases(0)
        self._kill_all_other_connections_to_the_database()

        return self.execute_command_in_proper_context(
            command='gunzip | ' + self._get_definition().get_restore_command(),
            mode='restore',
            stdin=stream,
            copy_stdin=True
        )

    def _finalize_restore(self):
        self.set_connection_limit_on_databases(-1)

    def set_connection_limit_on_databases(self, limit: int):
        cmd = self.execute_command_in_proper_context(self._get_definition().get_all_databases_command(), wait=300)
        databases = cmd.stdout.read().decode('utf-8').split("\n")

        for row in databases:
            database_name = row.strip()

            if not database_name:
                continue

            self.execute_command_in_proper_context(
                self._get_definition().get_connection_limit_setter_command(database_name, limit),
                wait=300
            )

    def _kill_all_other_connections_to_the_database(self):
        """
        To avoid error such as 'ERROR:  database "rojava" is being accessed by other users' we need to kick all
        users from the system before importing the data
        :return:
        """

        self.execute_command_in_proper_context(self._get_definition().terminate_all_sessions_command(), wait=300)


class PostgreSQLBaseBackup(AbstractDockerAwareHandler):
    """
<sphinx>
postgres_base
-------------

Performs a backup using replication protocol using tool "pg_basebackup".
Restore simply unpacks files in PostgreSQL data directory.

On restore:

    - Turns off the database
    - Moves renames old data directory (to keep previous version)
    - Creates new data directory from backup archive
    - Starts the database
</sphinx>
    """

    def _get_definition(self) -> PostgreSQLBaseBackupDefinition:
        return self._definition

    def receive_backup_stream(self):
        """ Do a copy using pg_basebackup """

        # ensure the temporary directory is empty and exists
        self.shell(self._get_definition().get_make_tempdir_command())

        return self.execute_command_in_proper_context(
            command=self._get_definition().get_backup_command() + ' | gzip',
            mode='backup'
        )

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        """ Restore backup on an offline database """
        self._shutdown_the_database()
        self._rename_current_data_dir()

        return self.execute_command_in_proper_context(
            command='gunzip | ' + self._get_definition().get_restore_command(),
            mode='backup',
            stdin=stream,
            copy_stdin=True
        )

    def _rename_current_data_dir(self):
        self.shell(self._get_definition().get_rename_data_dir_command())

    def _shutdown_the_database(self):
        self.shell(self._get_definition().get_server_shutdown_command())

    def _start_the_database(self):
        self.shell(self._get_definition().get_server_start_command())

    def _set_permissions(self):
        self.shell(self._get_definition().get_permissions_restore_command())

    def on_failed_restore(self):
        self.shell(self._get_definition().get_rescue_command_on_failed_restore())

    def _finalize_restore(self):
        self._set_permissions()
        self._start_the_database()

    def _finalize_backup(self):
        # clean up the temporary directory
        self.shell(self._get_definition().get_temporary_directory_clean_up_command())




from .abstractdocker import AbstractDockerAwareHandler
from ..result import CommandExecutionResult
from ..entity.definition.local import PathBackupDefinition
import os


class FileOrDirectoryBackup(AbstractDockerAwareHandler):
    """
<sphinx>
path
----

Packs all specified files and/or directories when doing backup, then unpacks on doing restore.
Uses TAR with gzip compression. Supports docker container paths and host paths.

**Examples:**

.. code::yaml

    some_local_dir:
        type: path
        paths:
            - /tmp/test
        access: backup_one
        encryption: enc1
        collection_id: "${COLLECTION_ID}"

        # optional
        #tar_pack_cmd: "tar -czf %stdin% %paths%"
        #tar_unpack_cmd: "tar -xzf %stdin% %target%"

    dir_in_docker:
        type: path
        container: riotkit_storage_1     # notice the "container" there
        paths:
            - /var/storage
            - /var/db/storage-db.sqlite3
        access: backup_one
        encryption: enc1
        collection_id: "${COLLECTION_ID}"

        # optional
        #tar_pack_cmd: "tar -czf %stdin% %paths%"
        #tar_unpack_cmd: "tar -xzf %stdin% %target%"

</sphinx>

    """

    def _get_definition(self) -> PathBackupDefinition:
        return self._definition

    def validate_before_creating_backup(self):
        if self.is_using_container():
            definition = self._get_definition()

            for path in definition.get_paths():
                self._assert_path_exists_in_container(
                    path,
                    definition.get_docker_bin(),
                    self.get_container_name(),
                    definition
                )
            return

        for path in self._get_definition().get_paths():
            if not os.path.exists(path):
                raise Exception('Path "' + path + '" does not exist')

    def receive_backup_stream(self, container: str = None) -> CommandExecutionResult:
        """ Read from directory/file and return as a TAR-OPENSSL stream """

        tar_cmd = self._get_definition().get_pack_cmd(self._get_definition().get_paths())

        return self.execute_command_in_proper_context(
            command=tar_cmd,
            mode='backup',
            with_crypto_support=True,
            container=container
        )

    def restore_backup_from_stream(self, stream, container: str = None) -> CommandExecutionResult:
        """ Write to a directory/file - unpack a TAR archive """

        return self.execute_command_in_proper_context(
            command=self._get_definition().get_unpack_cmd(),
            mode='restore',
            with_crypto_support=True,
            stdin=stream,
            container=container,
            copy_stdin=True
        )



from ..entity.definition import BackupDefinition
from ..service.client import FileRepositoryClient
from ..service.pipefactory import PipeFactory
from ..exceptions import ReadWriteException
from ..result import CommandExecutionResult
from subprocess import TimeoutExpired
from logging import Logger
import string
import random
import subprocess
from shutil import copyfileobj


class BackupHandler:
    """ Manages the process of backup and restore, interacts with different sources of backup data using adapters """

    _client = None  # type: FileRepositoryClient
    _pipe_factory = None  # type: PipeFactory
    _logger = None  # type: Logger
    _definition = None
    _max_process_wait_timeout: int

    def __init__(self,
                 _client: FileRepositoryClient,
                 _pipe_factory: PipeFactory,
                 _logger: Logger,
                 _definition: BackupDefinition):
        self._client = _client
        self._pipe_factory = _pipe_factory
        self._logger = _logger
        self._definition = _definition
        self._max_process_wait_timeout = 3600

    def perform_backup(self):
        """ Prepares the backup and sends to the server """

        self.validate_before_creating_backup()

        self._logger.info('Receiving backup stream')
        response = self.receive_backup_stream()

        try:
            self._logger.info('Starting sending backup stream to server')

            # read first bytes, wait for process to early exit (if it would exit due to error)
            # that would allow us to grab its error code
            try:
                response.process.wait(3)
            except TimeoutExpired:
                pass

            self._raise_read_error_if_invalid_error_code(response)
            upload_response = self._client.send(response.process, self._get_definition())

        except Exception as e:
            response.process.kill()
            raise e

        finally:
            self.wait_for_process_to_finish(response.process)

        self._raise_read_error_if_invalid_error_code(response)
        return upload_response

    @staticmethod
    def _raise_read_error_if_invalid_error_code(response):
        if response.process.returncode is not None and response.process.returncode != 0:
            raise ReadWriteException('Backup source read error, use --debug and retry to investigate. ' +
                                     'Exit code: %i, Command: %s, Stderr: %s' %
                                     (response.process.returncode, response.command, response.stderr.read()[0:512]))

    def perform_restore(self, version: str):
        """ Downloads backup version from server and applies locally """

        response = None

        try:
            # @todo: Implement a switch, that would allow to save file to temporary path instead of piping
            #        It would be just an additional option for safety
            #        See: https://github.com/riotkit-org/file-repository/issues/101

            stream = self._read_from_storage(version)
            response = self.restore_backup_from_stream(stream)

            self.wait_for_process_to_finish(response.process)

            if response.process.returncode > 0:
                raise ReadWriteException('Cannot restore backup. Errors: '
                                         + str(response.stderr.read().decode('utf-8')))
        except Exception:
            self._logger.error('Executing on_failed_restore() if defined')
            self.on_failed_restore()

            if response:
                response.process.kill()

            raise

        self._logger.info('No errors found, sending success information')

        return '{"status": "OK"}'

    def close(self, action: str):
        self._logger.info('Finishing the process')

        if action == 'backup':
            self._finalize_backup()
        elif action == 'restore':
            self._finalize_restore()

    def _get_definition(self) -> BackupDefinition:
        return self._definition

    def wait_for_process_to_finish(self, process):
        self._logger.info('Waiting for process %i to finish, timeout=%i' %
                          (process.pid, self._max_process_wait_timeout))

        process.wait(self._max_process_wait_timeout)
        process.stdout.close()

    def _execute_command(self, command: str, stdin=None,
                         copy_stdin: bool = False, wait: int = None) -> CommandExecutionResult:
        """
        Executes a command on local machine, returning stdout as a stream, and streaming in the stdin (optionally)
        """

        self._logger.debug('shell(' + command + ')')

        process = subprocess.Popen(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   stdin=subprocess.PIPE if stdin else None,
                                   executable='/bin/bash',
                                   shell=True)

        if stdin and copy_stdin:
            self._logger.info('Copying stdin to process')

            try:
                copyfileobj(stdin, process.stdin)
            except BrokenPipeError:
                self._logger.warning('Broken pipe happened, trying to kill process')
                process.kill()
                process.wait(15)

                raise ReadWriteException(
                    'Cannot write to process, broken pipe occurred, probably a tar process died. '
                    + process.stderr.read().decode('utf-8')[0:512]
                )
            except Exception as e:
                self._logger.warning('Unknown problem happened, trying to kill process')
                process.kill()
                process.wait(15)

                raise ReadWriteException(
                    'Cannot write to process, unknown error occurred: ' + str(e) + ', exiting... '
                    + process.stderr.read().decode('utf-8')[0:512]
                )

            process.stdin.close()

        if wait is not None:
            process.wait(wait)

        return CommandExecutionResult(command, process.stdout, process.stderr, process)

    def validate_before_creating_backup(self):
        raise Exception('validate_before_creating_backup() not implemented for handler')

    def receive_backup_stream(self) -> CommandExecutionResult:
        """ TAR output or file stream buffered from ANY source for example """
        raise Exception('receive_backup_stream() not implemented for handler')

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        """ A file stream or tar output be written into the storage. May be OpenSSL encoded, depends on definition """
        raise Exception('restore_backup_from_stream() not implemented for handler')

    def _read_from_storage(self, version: str):
        return self._client.fetch(version, self._get_definition())

    def on_failed_restore(self):
        """ When restore is failed. Allows to specify a recovery procedure """
        pass

    def _finalize_backup(self):
        pass

    def _finalize_restore(self):
        pass

    @staticmethod
    def generate_id(size: int = 6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))




import subprocess
from typing import IO


class CommandExecutionResult:
    stdout: IO
    stderr: IO
    process: subprocess.Popen
    command: str

    def __init__(self, command: str, stdout: IO, stderr: IO, process: subprocess.Popen):
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        self.process = process

    def __del__(self):
        self.process.kill()






from ..entity.recovery import RecoveryPlan
from . import AbstractController
from .restore import RestoreController


class RecoverFromDisasterController(AbstractController):

    def perform(self, plan_name: str):
        self._logger.info('Recovery from disaster using "' + plan_name + '" recovery plan')
        results = {True: [], False: []}

        plan = self._definition_factory.get_recovery_plan(plan_name)

        for name in plan.get_planned_definitions_to_recover_in_order():
            result = self._perform_recovery(name, plan)
            results[result].append(name)

        return {
            'recovered': results[True],
            'not_recovered': results[False]
        }

    def _perform_recovery(self, definition_name: str, plan: RecoveryPlan) -> bool:
        controller = RestoreController(
            self._definition_factory,
            self._logger,
            self._mapping,
            self._client,
            self._notifier
        )

        self._logger.info('Performing recovery of "' + definition_name + '"')

        try:
            controller.perform(definition_name, 'latest')
            return True

        except Exception as e:
            if not plan.on_error_should_continue():
                raise e

            return False





from ..handler.dockervolumebackup import DockerVolumeHotBackup, DockerVolumeBackup
from ..handler.fileordirectorybackup import FileOrDirectoryBackup
from ..handler.commandoutputbackup import CommandOutputBackup
from ..handler.mysqlbackup import MySQLBackup
from ..handler.postgresbackup import PostgreSQLDumpBackup, PostgreSQLBaseBackup
from ..handler.redisbackup import RedisBackup


class HandlersMapping:
    _mapping = {
        # databases
        'mysql': MySQLBackup,
        'postgres': PostgreSQLDumpBackup,
        'postgres_base': PostgreSQLBaseBackup,

        # commands
        'docker_output': CommandOutputBackup,
        'command_output': CommandOutputBackup,

        # files
        'directory': FileOrDirectoryBackup,
        'file': FileOrDirectoryBackup,
        'path': FileOrDirectoryBackup,
        'docker_hot_volumes': DockerVolumeHotBackup,
        'docker_volumes': DockerVolumeBackup,

        # kv
        'redis': RedisBackup
    }

    def get(self, name: str):
        """ Resolves "type" configuration key into object, on error throws KeyError """

        return self._mapping[name]

    def has_handler(self, name: str) -> bool:
        return name in self._mapping



#!/usr/bin/env python

from copy import copy
import logging
import datetime
import sys

MAPPING = {
    'DEBUG': 37,
    'INFO': 36,
    'WARNING': 33,
    'ERROR': 31,
    'CRITICAL': 41
}

PREFIX = '\033['
SUFFIX = '\033[0m'


class ColoredFormatter(logging.Formatter):

    def __init__(self, pattern):
        logging.Formatter.__init__(self, pattern)

    def format(self, record):
        colored_record = copy(record)
        level_name = colored_record.levelname

        seq = MAPPING.get(level_name, 37)
        colored_level_name = '{0}{1}m{2}{3}' \
            .format(PREFIX, seq, level_name, SUFFIX)
        colored_record.levelname = colored_level_name

        return logging.Formatter.format(self, colored_record)


class LoggerFactory:

    @staticmethod
    def create(is_debug: bool, path: str, single_log_file: str) -> logging.Logger:
        """ Creates a logger instance with proper handlers configured """

        logger = logging.getLogger('bahub')
        logger.setLevel(logging.INFO)
        formatter = ColoredFormatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s")

        logging_handler = logging.StreamHandler(sys.stdout)
        logging_handler.setLevel(logging.INFO)
        logging_handler.setFormatter(formatter)

        log_file_name = 'bahub-{date:%Y-%m-%d}.log'.format(date=datetime.datetime.now())

        if single_log_file:
            log_file_name = single_log_file

        log_file_handler = logging.FileHandler(path + '/' + log_file_name, 'a+')
        log_file_handler.setFormatter(formatter)
        log_file_handler.setLevel(logging.INFO)

        if is_debug:
            logger.setLevel(logging.DEBUG)
            logging_handler.setLevel(logging.DEBUG)
            log_file_handler.setLevel(logging.DEBUG)
        else:
            logging_handler.setLevel(logging.ERROR)

        logger.addHandler(logging_handler)
        logger.addHandler(log_file_handler)

        return logger


class PasswordsProtectedFilter(logging.Filter):
    def __init__(self, patterns):
        super(PasswordsProtectedFilter, self).__init__()
        self._patterns = patterns

    def filter(self, record):
        record.msg = self.redact(record.msg)
        if isinstance(record.args, dict):
            for k in record.args.keys():
                record.args[k] = self.redact(record.args[k])
        else:
            record.args = tuple(self.redact(arg) for arg in record.args)
        return True

    def redact(self, msg):
        msg = str(msg)

        for pattern in self._patterns:
            if len(pattern) < 2:
                continue

            replacement = pattern[0]
            replacement += "*" * (len(pattern) - 2)
            replacement += pattern[-1:]

            msg = msg.replace(pattern, replacement)

        return msg





from .abstractdocker import AbstractDockerAwareHandler
from ..result import CommandExecutionResult
from ..entity.definition.sql import MySQLDefinition


class MySQLBackup(AbstractDockerAwareHandler):
    def _get_definition(self) -> MySQLDefinition:
        return self._definition

    def validate_before_creating_backup(self):
        pass

    def receive_backup_stream(self):
        """
        Use MySQL dump command to extract data from database
        """
        return self.execute_command_in_proper_context(
            command=self._get_definition().get_dump_command() + ' | gzip',
            mode='backup'
        )

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        """
        Use MySQL shell util to import the database again
        """

        self._logger.info('Not sending any DROP TABLE, mysqldump should already have "drop if exists" in dump')

        return self.execute_command_in_proper_context(
            command='gunzip | ' + self._get_definition().get_restore_command(),
            mode='restore',
            stdin=stream,
            copy_stdin=True
        )




from . import ContainerizedDefinition
from . import BackupDefinition
from ..access import ServerAccess
from ..encryption import Encryption


class DockerVolumesDefinition(ContainerizedDefinition):
    _paths: list

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 container: str,
                 docker_bin: str,
                 paths: list,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._container = container
        self._docker_bin = docker_bin
        self._paths = paths

    def get_paths(self) -> list:
        return self._paths

    @staticmethod
    def from_config(config: dict, name: str):
        return DockerVolumesDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            container=config['container'],
            docker_bin=config.get('docker_bin', 'docker'),
            paths=config['paths'],
            name=name
        )

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + ',docker_container=' + \
               str(self.get_container()) + '>'


class DockerOfflineVolumesDefinition(DockerVolumesDefinition):
    _paths = ''
    _temp_image_name = ''
    _temp_cmd = ''

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 container: str,
                 docker_bin: str,
                 paths: list,
                 temp_image_name: str,
                 temp_cmd: str,
                 name: str):
        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd,
                         tar_unpack_cmd, container, docker_bin, paths, name)

        self._container = container
        self._docker_bin = docker_bin
        self._paths = paths
        self._temp_image_name = temp_image_name
        self._temp_cmd = temp_cmd

    def get_temp_image_name(self):
        return self._temp_image_name

    def get_temp_cmd(self):
        return self._temp_cmd

    @staticmethod
    def from_config(config: dict, name: str):
        return DockerOfflineVolumesDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            container=config['container'],
            docker_bin=config.get('docker_bin', 'docker'),
            paths=config['paths'],
            temp_image_name=config.get('temp_image_name', 'alpine:3.9'),
            temp_cmd=config.get('temp_image_cmd', 'apk add --update xz bzip2 && sleep 3600'),
            name=name
        )




from .fileordirectorybackup import FileOrDirectoryBackup
from ..entity.definition.docker import DockerOfflineVolumesDefinition
from ..result import CommandExecutionResult


class DockerVolumeHotBackup(FileOrDirectoryBackup):
    """
<sphinx>
docker_hot_volumes
------------------

Alias to "path" with enforced container parameter.

Backups a RUNNING container. For some applications it may be safe to backup a running container, for some not.
See DockerVolumesBackup and adjust backup method to the application you want to keep safe.
Gets into the container and makes a backup of directories into a single tar file.

**Example:**

.. code:: yaml

    docker_hot_volumes_example:
        type: docker_hot_volumes
        container: "test_www"
        access: backup_one
        encryption: none
        collection_id: "${COLLECTION_ID}"
        paths:
            - /var/www
            - /var/log/nginx/access.log

        # optional
        #tar_pack_cmd: "tar -czf %stdin% %paths%"
        #tar_unpack_cmd: "tar -xzf %stdin% %target%"
        #docker_bin: "sudo docker"

</sphinx>
    """

    def is_using_container(self) -> bool:
        return True


class DockerVolumeBackup(DockerVolumeHotBackup):
    """
docker_volumes
--------------

Offline docker container backup. Runs a new container, mounts volumes of origin container and performs a backup
of those mounted volumes. Fully secure option for all kind of applications, as the applications are shut down for
a moment.

Notice: Your applications will have a downtime. Be careful about dependent services, those may exit unexpectedly.

**Example:**

.. code:: yaml

    www_docker_offline:
        type: docker_volumes
        container: "test_www"
        access: backup_one
        encryption: enc1
        collection_id: "${COLLECTION_ID}"
        paths:
            - /etc
            - /var/lib/mysql
            - /var/log/mysql.log

        # optional
        docker_bin: "docker"
        tar_pack_cmd: "tar -czf %stdin% %paths%"
        tar_unpack_cmd: "tar -xzf %stdin% %target%"
        temp_image_name: "alpine:3.9"
        temp_image_cmd: "apk add --update xz bzip2 && sleep 3600"

    """

    _temp_container_id = ""

    def _get_definition(self) -> DockerOfflineVolumesDefinition:
        return self._definition

    def validate_before_creating_backup(self):
        # do not check if container is up, it does not need to be. It must only exists.
        pass

    def receive_backup_stream(self, container: str = None):
        self._logger.info('Performing backup of origin container in offline mode')

        temporary_container_id = self._stop_origin_and_start_temporary_containers(
            image=self._get_definition().get_temp_image_name(),
            cmd=self._get_definition().get_temp_cmd()
        )

        return super().receive_backup_stream(container=temporary_container_id)

    def restore_backup_from_stream(self, stream, container: str = None) -> CommandExecutionResult:
        self._logger.info('Restoring backup to the temporary container through mounted volumes of origin container')

        temporary_container_id = self._stop_origin_and_start_temporary_containers(
            image=self._get_definition().get_temp_image_name(),
            cmd=self._get_definition().get_temp_cmd()
        )

        return super().restore_backup_from_stream(stream, container=temporary_container_id)

    def backup_container_directories(self, docker_bin: str, container: str, paths: list,
                                     definition: DockerOfflineVolumesDefinition) -> CommandExecutionResult:
        """ Performs a backup of multiple directories using TAR with gzip/xz/bz2 compression """

        return self._execute_in_container(
            docker_bin, container,
            definition.get_pack_cmd(paths),
            definition
        )

    def _finalize_backup(self):
        self._stop_temporary_and_start_origin_container()

    def _finalize_restore(self):
        self._stop_temporary_and_start_origin_container()




from ..service.configurationfactory import ConfigurationFactory
from ..entity.definition import BackupDefinition
from ..mapping.handlers import HandlersMapping
from ..handler import BackupHandler
from ..service.client import FileRepositoryClient
from ..service.pipefactory import PipeFactory
from ..service.notifier import NotifierInterface
from logging import Logger


class AbstractController:
    _definition_factory = None  # type: ConfigurationFactory
    _logger = None              # type: Logger
    _mapping = None             # type: HandlersMapping
    _client = None              # type: FileRepositoryClient
    _notifier = None            # type: NotifierInterface

    def __init__(self, factory: ConfigurationFactory, logger: Logger,
                 mapping: HandlersMapping, client: FileRepositoryClient,
                 notifier: NotifierInterface):
        self._definition_factory = factory
        self._logger = logger
        self._mapping = mapping
        self._client = client
        self._notifier = notifier

    def _init_handler(self, definition: BackupDefinition) -> BackupHandler:
        if not self._mapping.has_handler(definition.get_type()):
            raise Exception('Unknown type "' + definition.get_type() + '"')

        self._logger.debug('Using handler "%s"' % str(self._mapping.get(definition.get_type())))

        return self._mapping.get(definition.get_type())(
            _client=self._client,
            _pipe_factory=PipeFactory(),
            _logger=self._logger,
            _definition=definition
        )




from . import ContainerizedDefinition
from ..access import ServerAccess
from ..encryption import Encryption


class CommandOutputDefinition(ContainerizedDefinition):
    _command: str
    _restore_command: str

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 command: str,
                 restore_command: str,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._command = command
        self._restore_command = restore_command

    def get_command(self) -> str:
        return self._command

    def get_restore_command(self) -> str:
        return self._restore_command

    @staticmethod
    def from_config(config: dict, name: str):
        definition = CommandOutputDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', ContainerizedDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', ContainerizedDefinition._tar_unpack_cmd),
            command=config['command'],
            restore_command=config.get('restore_command', ''),
            name=name
        )

        definition._container = config.get('container', '')
        definition._docker_bin = config.get('docker_bin', 'docker')

        return definition


class PathBackupDefinition(ContainerizedDefinition):
    _paths = ""

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 paths: list,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._paths = paths

    def get_paths(self) -> list:
        return self._paths

    @staticmethod
    def from_config(config: dict, name: str):
        definition = PathBackupDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', ContainerizedDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', ContainerizedDefinition._tar_unpack_cmd),
            paths=config['paths'],
            name=name
        )

        definition._container = config.get('container', '')
        definition._docker_bin = config.get('docker_bin', 'docker')

        return definition




from . import ContainerizedDefinition


class RedisDefinition(ContainerizedDefinition):
    """
        Support for Redis KV store export
    """

    data_dir: str              # if not chosen, then will be auto-detected using redis-cli
    is_aof: bool               # if not chosen, will be auto-detected
    fix_aof_at_restore: bool
    restart_command: str       # if not chosen, then will be fallen back to systemd or docker restart

    @staticmethod
    def from_config(config: dict, name: str):
        definition = RedisDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', ContainerizedDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', ContainerizedDefinition._tar_unpack_cmd),
            name=name
        )

        definition._container = config.get('container', '')
        definition._docker_bin = config.get('docker_bin', 'docker')
        definition.data_dir = config.get('data_dir', None)
        definition.is_aof = config.get('is_aof', None)
        definition.fix_aof_at_restore = config.get('fix_aof_at_restore', False)
        definition.restart_command = config.get('restart_command', None)

        return definition






from ..entity.definition import BackupDefinition
from . import AbstractController


class ListController(AbstractController):
    """ Lists all stored backups for given backup slot """

    def do_ls(self, definition: BackupDefinition):
        return self._client.list_backups(definition)



from ..access import ServerAccess
from ..encryption import Encryption


class BackupDefinition:
    """ Configuration definition containing credentials, tokens, ids """

    _access = None       # type: ServerAccess
    _type = ""           # type: str
    _encryption = None   # type: Encryption
    _collection_id = ""  # type: str
    _tar_pack_cmd = 'tar -czf %stdin% %paths%'
    _tar_unpack_cmd = 'tar -xzf %stdin% %target%'
    _name = ""

    def __init__(self, access: ServerAccess, _type: str, collection_id: str, encryption: Encryption,
                 tar_pack_cmd: str, tar_unpack_cmd: str, name: str):
        self._access = access
        self._type = _type
        self._encryption = encryption
        self._collection_id = collection_id
        self._tar_pack_cmd = tar_pack_cmd
        self._tar_unpack_cmd = tar_unpack_cmd
        self._name = name

    @staticmethod
    def from_config(config: dict, name: str):
        return BackupDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            name=name
        )

    def get_pack_cmd(self, paths: list):
        return self._tar_pack_cmd \
            .replace('%stdin%', '-') \
            .replace('%paths%', self.join_paths_quotes_into_string(paths))

    def get_unpack_cmd(self):
        return self._tar_unpack_cmd \
            .replace('%target%', '-C /') \
            .replace('%stdin%', '-')

    @staticmethod
    def join_paths_quotes_into_string(paths: list):
        joined = ""

        for path in paths:
            joined += '"' + path + '" '

        return joined

    def get_access(self) -> ServerAccess:
        return self._access

    def get_encryption(self) -> Encryption:
        return self._encryption

    def get_type(self) -> str:
        return self._type

    def get_collection_id(self) -> str:
        return self._collection_id

    def get_sensitive_information(self) -> list:
        return []

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + '>'


class ContainerizedDefinition(BackupDefinition):
    _container: str
    _docker_bin: str

    def __init__(self, *args, **kwargs):
        self._container = ''
        self._docker_bin = 'docker'

        super().__init__(*args, **kwargs)

    def get_container(self) -> str:
        return self._container

    def get_docker_bin(self) -> str:
        return self._docker_bin

    def should_use_docker(self) -> bool:
        return self.get_container() != '' and self.get_docker_bin() != ''





from ..entity.encryption import Encryption
from ..entity.recovery import RecoveryPlan
from ..entity.access import ServerAccess
from ..entity.definition import BackupDefinition
from .errorhandler import ErrorHandlerFactory, ErrorHandlerInterface
from .notifier import NotifierInterface, NotifierFactory
from ..mapping.definitions import DefinitionsMapping
from ..exceptions import ConfigurationFactoryException, ConfigurationError
import yaml
from yaml import SafeLoader as Loader
import os
import re


class ConfigurationFactory:
    """ Constructs objects basing on the configuration file """

    _accesses = {}        # type: dict[ServerAccess]
    _encryption = {}      # type: dict[Encryption]
    _backups = {}         # type: dict[BackupDefinition]
    _recovery_plans = {}  # type: dict[RecoveryPlan]
    _error_handlers = {}  # type: dict[ErrorHandlerInterface]
    _notifiers = {}       # type: dict[NotifierInterface]
    _debug = False        # type: bool
    _config_dir: str

    def __init__(self, configuration_path: str, debug: bool):
        self._debug = debug
        self._config_dir = os.path.abspath(os.path.dirname(configuration_path))
        self._parse(self._read(configuration_path))

    def _parse(self, config: dict):
        self._parse_accesses(config['accesses'])
        self._parse_encryption(config['encryption'])
        self._parse_backups(config['backups'])
        self._parse_monitoring_error_handlers(config.get('error_handlers', {}))
        self._parse_notifiers(config.get('notifiers', {}))

        # recovery plans are optional
        self._parse_recovery_plans(config['recoveries'] if 'recoveries' in config else {})

    def _read(self, path: str):
        f = open(path, 'rb')

        try:
            config = yaml.load(self._process_env_variables(f.read().decode('utf-8')), Loader=Loader)
        except Exception as e:
            f.close()
            raise e

        f.close()
        return config

    def _process_env_variables(self, content: str) -> str:
        env_list = list(dict(os.environ).items())
        env_list.sort(key=lambda item: (-len(item[0]), item[0]))

        # add special variables:
        env_list.append(['CONFIG_DIR', self._config_dir])

        for env in env_list:
            content = content.replace('${' + env[0] + '}', env[1])

        invalid_vars = set(re.findall('\${([A-Z0-9a-z_]+)}', content))

        if len(invalid_vars) > 0:
            raise ConfigurationFactoryException(
                'Following environment variables are not resolved: ' + (', '.join(invalid_vars))
            )

        return content

    def _parse_accesses(self, config: dict):
        """ Access tokens """

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('accesses.' + key, self._debug):
                self._accesses[key] = ServerAccess.from_config(values)

    def _parse_encryption(self, config: dict):
        """ Security/Encryption """

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('encryption.' + key, self._debug):
                self._encryption[key] = Encryption.from_config(values)

    def _parse_recovery_plans(self, config: dict):
        """ Recovery plans/strategies """
        possible_backup_definitions = list(self._backups.keys())

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('recoveries.' + key, self._debug):
                self._recovery_plans[key] = RecoveryPlan.from_config(values, possible_backup_definitions)

    def _parse_backups(self, config: dict):
        """ Backups """

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('backups.' + key, self._debug):

                # find related access and encryption
                values['access'] = self._accesses[values['access']]

                if "encryption" in values:
                    values['encryption'] = self._encryption[values['encryption']]

                factory_method = DefinitionsMapping.get(values['type'])

                try:
                    self._backups[key] = factory_method.from_config(values, key)

                except KeyError as config_key_name:
                    raise ConfigurationError('Backup "%s" is missing "%s" configuration option' %
                                             (key, config_key_name))

    def _parse_monitoring_error_handlers(self, config: dict):
        """ Error handlers integration """

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('error_handlers.' + key, self._debug):

                if 'type' not in values:
                    raise ConfigurationFactoryException('Error handler must have a defined type')

                self._error_handlers[key] = ErrorHandlerFactory.create(values['type'], values)

    def _parse_notifiers(self, config: dict):
        """ Notifiers """

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('notificers..' + key, self._debug):

                if 'type' not in values:
                    raise ConfigurationFactoryException('Notifier type needs to be specified')

                self._notifiers[key] = NotifierFactory.create(values['type'], values)

    def get_error_handlers(self):
        return self._error_handlers

    def get_notifiers(self):
        return self._notifiers

    def get_definition(self, name: str) -> BackupDefinition:
        if name not in self._backups:
            raise ConfigurationFactoryException(
                'No such backup definition, maybe a typo? Please check the configuration file'
            )

        return self._backups[name]

    def get_recovery_plan(self, name: str) -> RecoveryPlan:
        if name not in self._recovery_plans:
            raise ConfigurationFactoryException(
                'Specified recovery plan not found'
            )

        return self._recovery_plans[name]

    def get_all_sensitive_data(self) -> list:
        sensitive_data = []

        for access in self._accesses:
            sensitive_data.append(self._accesses[access].get_token())

        for backup in self._backups:
            if self._backups[backup].get_encryption().get_passphrase():
                sensitive_data.append(self._backups[backup].get_encryption().get_passphrase())

            sensitive_data.append(self._backups[backup].get_collection_id())
            sensitive_data += self._backups[backup].get_sensitive_information()

        return sensitive_data


class DefinitionFactoryErrorCatcher:
    _key_name = ""
    _debug = False

    def __init__(self, config_key_name: str, _debug: bool):
        self._key_name = config_key_name
        self._debug = _debug

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            if self._debug:
                return

            raise ConfigurationFactoryException(
                ' ERROR: There was a problem during parsing the configuration at section "' +
                self._key_name + '" in key ' + str(exc_val) + ', details: ' + str(exc_type))





from .abstractdocker import AbstractDockerAwareHandler
from ..result import CommandExecutionResult
from ..exceptions import ReadWriteException
from ..entity.definition.local import CommandOutputDefinition


class CommandOutputBackup(AbstractDockerAwareHandler):
    """
<sphinx>
command_output
--------------

Grabs output of a command. Runs on host and in docker.
Possibility of using the restore depends on well defined "restore_command" parameter.

**Example:**

.. code:: yaml

    local_command_output:
        type: command_output
        command: "cat /bin/bash"
        container: "some_container"                    # optional
        restore_command: "cat - > /tmp/bash.restored"  # optional
        access: backup_one
        encryption: enc1
        collection_id: "${COLLECTION_ID}"
</sphinx>

    """

    def _get_definition(self) -> CommandOutputDefinition:
        return self._definition

    def receive_backup_stream(self) -> CommandExecutionResult:
        return self.execute_command_in_proper_context(
            command=self._get_definition().get_command(),
            mode='backup',
            with_crypto_support=True
        )

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        definition = self._get_definition()

        if not definition.get_restore_command():
            raise ReadWriteException('Restore command not defined, cannot restore')

        return self.execute_command_in_proper_context(
            command=definition.get_restore_command(),
            mode='restore',
            with_crypto_support=True,
            stdin=stream,
            copy_stdin=True
        )




from .abstractdocker import AbstractDockerAwareHandler
from ..result import CommandExecutionResult
from ..entity.definition.kv import RedisDefinition


class RedisBackup(AbstractDockerAwareHandler):
    """
<sphinx>
redis
-----

Performs a backup of a REDIS server in a natural way.

Restore mechanism:

    1. Restore REDIS data files
    2. Optionally fix AOF log if needed
    3. Restart the service

Backup mechanism:

    1. Write all pending changes in REDIS to disk
    2. Grab the data from the disk

**Example:**

.. code:: yaml

    local_command_output:
        type: redis
        data_dir: "/data"                # optional, leave for auto-detection
        aof: true                        # optional, backup mode: AOF logs / NON-AOF (leave for auto-detection)
        fix_aof_at_restore: true         # optional, attempt to fix corrupted AOF before restoring? (could take time)
        container: "some_container"      # optional
        access: backup_one
        encryption: enc1
        collection_id: "${COLLECTION_ID}"
</sphinx>

    """

    def _get_definition(self) -> RedisDefinition:
        return self._definition

    def receive_backup_stream(self) -> CommandExecutionResult:
        """
            Write REDIS state to disk, then copy files
        """

        return self.execute_command_in_proper_context(
            command='redis-cli "SAVE" > /dev/null && ' + self._get_definition().get_pack_cmd([self.get_data_dir()]),
            mode='backup',
            with_crypto_support=True
        )

    def restore_backup_from_stream(self, stream) -> CommandExecutionResult:
        """
            Restore REDIS data files, then optionally fix AOF log if needed, restart the service
        """

        definition = self._get_definition()
        container_id = None

        if self.is_using_container():
            container_id = self._stop_origin_and_start_temporary_containers(image='', cmd='sleep 7200')

        return self.execute_command_in_proper_context(
            command=definition.get_unpack_cmd(),
            mode='restore',
            with_crypto_support=True,
            stdin=stream,
            copy_stdin=True,
            container=container_id
        )

    def _finalize_restore(self):
        """
            Docker: Stop temporary container and start the right one
            Bare metal: Restart the REDIS service
        """

        if self.is_fixing_aof_at_restore():
            self.fix_aof_logs()

        if self.is_using_container():
            self._stop_temporary_and_start_origin_container()
            return

        # bare metal
        self.shell(self.get_restart_command())

    def is_fixing_aof_at_restore(self) -> bool:
        return self.is_restoring_from_logs() and self._get_definition().fix_aof_at_restore

    def fix_aof_logs(self):
        """
            Attempt to fix corrupted logs (if they are) by running REDIS tool "redis-check-aof" before starting server.
            This guarantees, that the recovered backup will work. The fixing procedure is optional.

            https://redis.io/topics/persistence
        """

        self._logger.info('Running REDIS AOF fixer before starting server to make sure the server will start even ' +
                          'if logs would be corrupted')

        fix_cmd = 'redis-check-aof --fix ' + self.get_data_dir() + '/appendonly.aof'

        if self.is_using_container():
            self.execute_command_in_proper_context(
                command=fix_cmd,
                container=self._temp_container_id,
                wait=7200
            )
            return

        self.shell(fix_cmd, wait=7200)

    def get_data_dir(self) -> str:
        if self._get_definition().data_dir:
            return self._get_definition().data_dir

        # automatic detection
        return self.shell('redis-cli --raw config get dir|tail -n -').stdout.read().decode('utf-8').strip()

    def is_restoring_from_logs(self) -> bool:
        """
            Checks for AOF setting - https://redis.io/topics/persistence
        """

        if self._get_definition().is_aof is not None:
            return self._get_definition().is_aof

        # automatic detection
        return self.shell('redis-cli --raw config get appendonly|tail -n -1').\
                   stdout.read().decode('utf-8').strip() == 'yes'

    def get_restart_command(self) -> str:
        """
            Restart command performed after restore - manually entered, or auto-detected:
            1) nothing if in docker, as we already setup temporary containers and doing the restart by the way
            2) systemd when not using docker container
        """

        if self._get_definition().restart_command is not None:
            return self._get_definition().restart_command

        if self.is_using_container():
            return ''

        return 'systemctl restart redis'




from . import BackupHandler
from ..entity.definition import ContainerizedDefinition
from ..exceptions import ReadWriteException
from ..result import CommandExecutionResult


class AbstractDockerAwareHandler(BackupHandler):
    """
        Base class for all handlers that uses docker
        When a handler inherits from this class it does not mean it works only with docker, can work with both
        host and docker (execute_command_in_proper_context() is example of such behavior)

        :author: RiotKit Team, Andrew Johnson
    """

    _temp_container_id: str

    def _get_definition(self) -> ContainerizedDefinition:
        return self._definition

    def validate_before_creating_backup(self):
        definition = self._get_definition()

        if self.is_using_container():
            self.assert_container_running(
                definition.get_docker_bin(),
                self.get_container_name()
            )

    def is_using_container(self) -> bool:
        return self._get_definition().should_use_docker()

    def _execute_in_container(self, docker_bin: str, container: str, command: str, definition: ContainerizedDefinition,
                              allocate_pts=False,
                              interactive=False,
                              with_crypto=True,
                              stdin=None,
                              mode='backup',
                              wait: int = None,
                              copy_stdin: bool = False) -> CommandExecutionResult:

        """ Executes any command inside of the container """

        opts = ''

        if allocate_pts:
            opts += ' -t'

        if interactive:
            opts += ' -i'

        method = self._pipe_factory.create_pure_command

        # the pipe factory constructs the end command, including details such as encryption
        if mode == 'backup':
            method = self._pipe_factory.create_backup_command

        if mode == 'restore':
            method = self._pipe_factory.create_restore_command

        return self._execute_command(
            method(
                docker_bin + ' exec ' + opts + ' ' + container + ' /bin/sh -c "' + command.replace('"', '\\"') + '"',
                definition,
                with_crypto=with_crypto
            ),
            stdin=stdin,
            copy_stdin=copy_stdin,
            wait=wait
        )

    def execute_command_in_proper_context(self, command: str, mode: str = '', with_crypto_support: bool = True,
                                          stdin=None, container: str = None,
                                          wait: int = None, copy_stdin: bool = False) -> CommandExecutionResult:
        """
        Execute command in docker or on host
        """

        factory_method = self._pipe_factory.create_pure_command

        if mode == 'restore':
            factory_method = self._pipe_factory.create_restore_command

        elif mode == 'backup':
            factory_method = self._pipe_factory.create_backup_command

        definition = self._get_definition()

        if self.is_using_container():
            return self._execute_in_container(
                definition.get_docker_bin(),
                container if container else self.get_container_name(),
                command,
                definition,
                mode=mode,
                interactive=True,
                stdin=stdin,
                copy_stdin=copy_stdin,
                wait=wait
            )

        return self._execute_command(
            factory_method(
                command=command,
                definition=self._get_definition(),
                with_crypto=with_crypto_support
            ),
            stdin=stdin,
            copy_stdin=copy_stdin,
            wait=wait
        )

    def shell(self, cmd: str, stdin=None, wait: int = 30):
        """ Shortcut for execute_command_in_proper_context() to just execute freely shell commands """

        return self.execute_command_in_proper_context(
            command=cmd,
            mode='',
            with_crypto_support=False,
            stdin=stdin,
            copy_stdin=stdin is not None,
            wait=wait
        )

    def get_container_name(self) -> str:
        return self._get_definition().get_container()

    def _spawn_temporary_container(self, docker_bin: str, origin_container: str, temp_image_name: str,
                                   temp_container_cmd: str):
        """
            Runs a temporary container that has mounted volumes from other container

            Notice: Empty "temp_image_name" parameter will mean to run same image as origin container
        """

        self._logger.info('Bringing up temporary container')

        temp_container_name = origin_container + '_backup_' + self.generate_id()

        if temp_image_name == '':
            temp_image_name = self._inspect_docker_container_image(docker_bin, origin_container)

        run_command = docker_bin + ' run -d --rm --volumes-from ' + origin_container + \
                                   ' --name ' + temp_container_name + \
                                   ' ' + temp_image_name + \
                                   ' /bin/sh -c "' + temp_container_cmd + '"'

        result = self._execute_command(
            self._pipe_factory.create_backup_command(
                run_command,
                self._get_definition(),
                with_crypto=False
            ),
            wait=300
        )

        out, code = result.process.communicate()
        container_id = out.decode('utf-8').strip()

        if not container_id or (result.process.returncode != 0 and result.process.returncode is not None):
            raise ReadWriteException('Cannot run temporary docker container, please verify image and command. Output: '
                                     + container_id)

        return container_id

    def _inspect_docker_container_image(self, docker_bin: str, container_name: str) -> str:
        return self.shell(docker_bin + ' inspect --format="{{.Image}}" ' + container_name).stdout.read().decode('utf-8').strip()

    def _assert_all_paths_exists(self, docker_bin: str, container: str, paths: list,
                                 definition: ContainerizedDefinition):
        """ Multiple assertion for directory/files presence """

        self.assert_container_running(docker_bin, container)

        for path in paths:
            self._assert_path_exists_in_container(docker_bin, container, path, definition)

    def _assert_path_exists_in_container(self,
                                         path: str,
                                         docker_bin: str,
                                         container: str,
                                         definition):
        """ Checks if a single directory or file exists inside of a container """

        response = self._execute_in_container(
            docker_bin,
            container,
            '[ -e ' + path + ' ] || echo does-not-exist',
            definition,
            with_crypto=False
        )

        code = response.process.wait()
        out = response.stdout.read().decode('utf-8') + response.stderr.read().decode('utf-8')

        if "does-not-exist" in out:
            raise ReadWriteException(
                'Path "' + path + '" does not exist in container "' + self.get_container_name() + '"'
            )

        if int(code) > 0:
            raise ReadWriteException('Non-zero exit code from command, check if the container name is valid. Out: ' +
                                     out)

    def assert_container_running(self, docker_bin: str, container: str):
        """ Checks if a docker container is running """

        response = self._execute_command(
            docker_bin + ' ps | grep "' + container + '"',
            wait=30
        )

        response.process.wait()
        output = (str(response.stdout.read()) + str(response.stderr.read())).lower()

        if "got permission denied while trying to connect" in output:
            raise ReadWriteException(
                'You do not have access rights to the docker daemon, ' +
                'shoudn\'t you use sudo in docker_bin or run bahub as root?'
            )

        if "cannot connect" in output or "connection refused" in output:
            raise ReadWriteException(
                'Docker daemon seems to be not running, or you do not have access rights to access it'
            )

        if container not in output:
            raise ReadWriteException('Container seems to be not running, check docker ps')

        if response.process.returncode > 0:
            raise ReadWriteException('Command failed with non-zero exit code, ' +
                                     'probably the container is not running, output: ' + output)

    @staticmethod
    def _assert_container_name_present(out: CommandExecutionResult, container_name: str, operation_name: str):
        stdout = out.stdout.read().decode('utf-8').replace('"', '').replace("'", '').strip()
        container_name = container_name.replace('"', '').strip()

        if not stdout == container_name:
            raise ReadWriteException('Cannot ' + operation_name + ' container "' + container_name + '". ' + stdout)

    def _kill_container(self, container_name: str):
        docker_bin = self._get_definition().get_docker_bin()

        out = self._execute_command(docker_bin + ' kill ' + container_name, wait=300)
        self._assert_container_name_present(out, container_name, 'kill')

    def _stop_container(self, container_name: str):
        docker_bin = self._get_definition().get_docker_bin()

        out = self._execute_command(docker_bin + ' stop ' + container_name, wait=300)
        self._assert_container_name_present(out, container_name, 'stop')

    def _start_container(self, container_name: str):
        docker_bin = self._get_definition().get_docker_bin()

        out = self._execute_command(docker_bin + ' start ' + container_name, wait=300)
        self._assert_container_name_present(out, container_name, 'start')

    def _stop_origin_and_start_temporary_containers(self, image: str, cmd: str) -> str:
        """ Stop origin container and start a temporary container """

        # @todo: Support linked/dependent containers

        definition = self._get_definition()

        self._logger.info('Stopping origin container')
        self._stop_container(definition.get_container())

        self._logger.info('Spawning temporary container with volumes from origin container')
        self._temp_container_id = self._spawn_temporary_container(
            definition.get_docker_bin(),
            definition.get_container(),
            image,
            cmd
        )

        return self._temp_container_id

    def _stop_temporary_and_start_origin_container(self):
        definition = self._get_definition()

        try:
            self._logger.info('Killing temporary container')
            self._kill_container(self._temp_container_id)

        except Exception:
            self._logger.warning('Cannot kill temporary container "' + self._temp_container_id + '"')

        # @todo: Support linked/dependent containers

        self._logger.info('Starting origin container')
        self._start_container(definition.get_container())




from ..entity.definition import BackupDefinition
from ..exceptions import InvalidRequestException, UnexpectedResponseException, ReadWriteException
from logging import Logger
from json import loads as json_loads
from json import JSONDecodeError
from typing import Union
import requests
import pycurl
import subprocess
import certifi
import io


class ProcessReader:
    """
    Reads from the opened UNIX process and reacts on errors while streaming
    In case when the process would exit it would raise an exception instead of ending the stream with EOF
    """

    process: subprocess.Popen
    exception: any  # type: Union[Exception, None]

    def __init__(self, process: subprocess.Popen):
        self.process = process
        self.exception = None

    def read_callback(self, size):
        try:
            self.process.wait(0)
        except subprocess.TimeoutExpired:
            pass

        if self.process.returncode is not None and self.process.returncode > 0:
            self.exception = ReadWriteException('Interrupting! The process exited early. Exit code: %i, Stderr: %s' %
                                                (self.process.returncode,
                                                 self.process.stderr.read().decode('utf-8')[0:512]))
            raise self.exception

        return self.process.stdout.read(size)


class FileRepositoryClient:
    """ Http Client to the File Repository API """

    _logger = None  # type: Logger

    def __init__(self, _logger: Logger):
        self._logger = _logger

    def send(self, process: subprocess.Popen, definition: BackupDefinition):
        """
        Sends the backup to the File Repository
        Supports exiting on interruption instead of sending corrupted data

        :param subprocess.Popen process:
        :param BackupDefinition definition:
        :return:
        """

        process_reader = ProcessReader(process)
        response_body_stream = io.BytesIO()
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup', True)

        curl = pycurl.Curl()
        curl.setopt(curl.URL, url)
        curl.setopt(curl.CAINFO, certifi.where())
        curl.setopt(curl.CUSTOMREQUEST, 'POST')
        curl.setopt(curl.UPLOAD, 1)
        curl.setopt(curl.READFUNCTION, process_reader.read_callback)
        curl.setopt(curl.WRITEFUNCTION, response_body_stream.write)
        curl.setopt(curl.VERBOSE, False)

        try:
            curl.perform()
        except pycurl.error as e:
            # raise the exception directly
            if process_reader.exception:
                raise process_reader.exception

            raise Exception('HTTP request error: ' + str(e) + '. Probably the application ' +
                            'backup process exited or timed out unexpectedly. Read above messages for details')

        response_body = response_body_stream.getvalue().decode('utf-8')

        self._logger.debug('Request: ' + str(url))
        self._logger.debug('response(' + response_body + ')')

        try:
            _json = json_loads(response_body)
        except JSONDecodeError:
            _json = {}

        if curl.getinfo(pycurl.HTTP_CODE) >= 400:
            raise InvalidRequestException(response_body, _json, _json.get('error_code', 0))

        curl.close()

        return {
            'version': _json['version']['version'],
            'file_id': _json['version']['id'],
            'file_name': _json['version']['file']['filename']
        }

    def fetch(self, version: str, definition: BackupDefinition):
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup/' + version,
            with_token=True)

        response = requests.get(url, stream=True)

        self._logger.debug('Request: ' + str(url))
        self._logger.debug('response_code=' + str(response.status_code))

        if response.status_code >= 400:
            self._logger.debug('response_code=' + str(response.text))
            raise InvalidRequestException(response.text, response.json(), response.json().get('error_code', 0))

        return response.raw

    def list_backups(self, definition: BackupDefinition) -> dict:
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup',
            with_token=True)

        response = requests.get(url)
        json = response.json()
        versions = dict()

        if response.status_code >= 400:
            raise InvalidRequestException(response.text, response.json(), response.json().get('error_code', 0))

        if 'versions' not in json:
            raise UnexpectedResponseException(response.text)

        for k, version in json['versions'].items():
            versions['v' + str(version['details']['version'])] = {
                'id': version['details']['id'],
                'created': version['details']['creation_date']['date']
            }

        return versions




from ..exceptions import ConfigurationError
from ..entity.definition import BackupDefinition
from ..entity.definition.sql import MySQLDefinition, PostgreSQLDefinition, PostgreSQLBaseBackupDefinition
from ..entity.definition.local import PathBackupDefinition, CommandOutputDefinition
from ..entity.definition.docker import DockerVolumesDefinition, DockerOfflineVolumesDefinition
from ..entity.definition.kv import RedisDefinition


class DefinitionsMapping:
    _mapping = {
        # databases
        'mysql': MySQLDefinition,
        'postgres': PostgreSQLDefinition,
        'postgres_base': PostgreSQLBaseBackupDefinition,

        # commands
        'docker_output': CommandOutputDefinition,
        'command_output': CommandOutputDefinition,
        'docker_hot_volumes': DockerVolumesDefinition,
        'docker_volumes': DockerOfflineVolumesDefinition,

        # files
        'directory': PathBackupDefinition,
        'file': PathBackupDefinition,
        'path': PathBackupDefinition,

        # kv
        'redis': RedisDefinition
    }

    @staticmethod
    def get(name: str) -> BackupDefinition:
        """ Resolves "type" configuration key into object, on error throws KeyError """

        if name not in DefinitionsMapping._mapping:
            raise ConfigurationError('%s is not a valid type of a backup')

        return DefinitionsMapping._mapping[name]




from . import ContainerizedDefinition
from . import BackupDefinition
from ..access import ServerAccess
from ..encryption import Encryption
from abc import ABC, abstractmethod


class AbstractSQLDefinition(ABC, ContainerizedDefinition):
    """ Common model for all SQL backups, contains connection details that are common across all databases """

    _host: str
    _port: int
    _user: str
    _password: str
    _database: str

    def __init__(self,
                 access: ServerAccess, _type: str, collection_id: str, encryption: Encryption,
                 tar_pack_cmd: str, tar_unpack_cmd: str, host: str, port: int, user: str,
                 password: str, database: str, docker_bin: str, container: str, name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)

        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._database = database
        self._docker_bin = docker_bin
        self._container = container

    def get_database(self) -> str:
        return self._database

    def fill_template(self, template: str):
        return template\
            .replace('%host%', self._host) \
            .replace('%user%', self._user) \
            .replace('%port%', str(self._port)) \
            .replace('%database%', self.get_database()) \
            .replace('%password%', self._password)

    def get_sensitive_information(self):
        return [self._password]


class AbstractDumpAndRestoreDefinition(AbstractSQLDefinition):
    """ Common model for dump & restore workflow when working with databases """

    _dump_cmd: str
    _restore_cmd: str
    _query_cmd: str

    _dump_opts: str
    _restore_opts: str

    def fill_template(self, template: str):
        return super().fill_template(template) \
            .replace('%dump_opts%', self._dump_opts) \
            .replace('%restore_opts%', self._restore_opts)

    @classmethod
    def from_config(cls, config: dict, name: str):
        definition = cls(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            host=config['host'],
            port=int(config['port']),
            user=config['user'],
            password=config['password'],
            database=config['database'],
            docker_bin=config.get('docker_bin', 'docker'),
            container=config.get('container', ''),
            name=name
        )

        definition._dump_opts = config.get('dump_opts', '')
        definition._restore_opts = config.get('restore_opts', '')

        definition._init_cmds(config)
        return definition

    def get_dump_command(self) -> str:
        return self.fill_template(self._dump_cmd)

    def get_restore_command(self):
        return self.fill_template(self._restore_cmd)

    @abstractmethod
    def _init_cmds(self, config: dict):
        raise NotImplementedError('Please implement _init_cmds() in child class')

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + \
               str(self.get_collection_id()) + ',docker_container=' + \
               str(self.get_container()) + ',sql_host=' + str(self._host) + '>'


class MySQLDefinition(AbstractDumpAndRestoreDefinition):
    """ MySQL backup model using mysqldump command """

    def _init_cmds(self, config: dict):
        self._dump_cmd = config.get('dump_cmd')
        self._restore_cmd = config.get('restore_cmd')

        if not self._dump_opts:
            self._dump_opts = ' --skip-lock-tables --add-drop-table --add-drop-database --add-drop-trigger '

        if not self._dump_cmd:
            self._dump_cmd = \
                'mysqldump %dump_opts% -u %user% -P %port% -p%password% ' + \
                '-h %host% --all-databases'

            if self.get_database():
                self._dump_cmd = \
                    'mysqldump %dump_opts% -u %user% -P %port% -p%password% ' + \
                    '-h %host% %database% '

        if not self._restore_cmd:
            self._restore_cmd = 'mysql -u %user% -p%password% -h %host% -P %port%'

            if self.get_database():
                self._restore_cmd = 'mysql -u %user% -p%password% -h %host% -P %port% %database%'

    def get_database(self) -> str:
        return str(self._database)

    def is_copying_all_databases(self) -> bool:
        return self.get_database() != ''


class PostgreSQLDefinition(AbstractDumpAndRestoreDefinition):
    """ PostgreSQL definition for using pg_dump pg_dumpall and pg_restore """

    def _init_cmds(self, config: dict):
        self._dump_cmd = config.get('dump_cmd')

        if not self._dump_cmd:
            #
            # Decide if we use pg_dump or pg_dumpall
            #
            if self.get_database():
                pg_dump_part = 'pg_dump -U %user% -p %port% -h %host% -d %database% -c %dump_opts%'
            else:
                pg_dump_part = 'pg_dumpall -U %user% -p %port% -h %host% -c %dump_opts%'

            self._dump_cmd = 'PGPASSWORD=%password% ' + pg_dump_part

        self._restore_cmd = config.get('restore_cmd')

        if not self._restore_cmd:
            db = self.get_database() if self.get_database() else 'postgres'
            self._restore_cmd = self.create_psql_cmdline(db)

    def terminate_all_sessions_command(self) -> str:
        return self.fill_template(
            'echo "SELECT pg_terminate_backend(pid) FROM pg_stat_activity  WHERE pid <> pg_backend_pid() "' +
            ' | ' + self.create_psql_cmdline('postgres')
        )

    def get_all_databases_command(self):
        return self.fill_template(
            'echo "SELECT datname FROM pg_database WHERE datistemplate = false;"' +
            ' | ' + self.create_psql_cmdline('postgres') + ' | sed \'/^$/d\''
        )

    @staticmethod
    def create_psql_cmdline(db: str = '') -> str:
        return 'PGPASSWORD=%password% psql -t -U %user% -p %port% -h %host% ' + db

    def get_connection_limit_setter_command(self, database_name: str, limit: int):
        return self.fill_template(
            'echo "ALTER DATABASE ' + database_name + ' CONNECTION LIMIT ' + str(limit) + ';"' +
            ' | ' + self.create_psql_cmdline('postgres')
        )


class PostgreSQLBaseBackupDefinition(AbstractSQLDefinition):
    """
        PostgreSQL backup using pg_basedump
    """

    _base_backup_cmd: str
    _server_shutdown_cmd: str
    _server_start_cmd: str
    _temp_dir: str
    _pack_cmd: str
    _restore_dir: str
    _ownership: str

    def fill_template(self, template: str):
        return super().fill_template(template) \
            .replace('%temp_dir%', self._temp_dir) \
            .replace('%restore_dir%', self._restore_dir) \
            .replace('%old_dir%', self._old_dir)

    def _init_cmds(self, config: dict):
        self._base_backup_cmd = config.get('base_backup_cmd')
        self._server_shutdown_cmd = config['server_shutdown_cmd']
        self._server_start_cmd = config['server_start_cmd']
        self._restore_dir = config['restore_dir'].rstrip('/')
        self._old_dir = config['old_dir']
        self._temp_dir = config.get('temp_dir', '/tmp/.pg_basebackup')
        self._pack_cmd = config.get('pack_cmd', 'cd %temp_dir% && tar --exclude=".." --exclude="." -cvf - * .*')
        self._ownership = config.get('ownership', '')

        if not self._base_backup_cmd:
            self._base_backup_cmd = 'PGPASSWORD=%password% ' + \
                                    'pg_basebackup -X stream -h %host% -U %user% -v -D %temp_dir% -F plain'

    def get_backup_command(self) -> str:
        """ Uses pg_basebackup to dump all required database files and configuration files """

        return self.fill_template(
            self._base_backup_cmd + ' && ' + self._pack_cmd
        )

    def get_restore_command(self) -> str:
        """ Will unpack tar archive with database files """

        return self.fill_template('tar --same-owner -xf - -C %restore_dir%')

    def get_rename_data_dir_command(self) -> str:
        """ We want to keep the previous data in {{ old_dir }} directory, for safety """

        return self.fill_template('rm -rf %old_dir% && mv %restore_dir% %old_dir% && mkdir -p %restore_dir%')

    def get_rescue_command_on_failed_restore(self) -> str:
        return self.fill_template('mv %restore_dir% %restore_dir%.abort && mv %old_dir% %restore_dir%')

    def get_server_shutdown_command(self) -> str:
        return self._server_shutdown_cmd

    def get_server_start_command(self) -> str:
        return self._server_start_cmd

    def get_permissions_restore_command(self):
        if self._ownership:
            return self.fill_template('chown -R ' + self._ownership + ' %restore_dir%')

        return ''

    def get_make_tempdir_command(self) -> str:
        return self.fill_template('rm -rf %temp_dir% && mkdir -p %temp_dir%')

    def get_temporary_directory_clean_up_command(self) -> str:
        return self.fill_template('rm -rf %temp_dir%')

    @classmethod
    def from_config(cls, config: dict, name: str):
        definition = cls(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            host=config['host'],
            port=int(config['port']),
            user=config['user'],
            password=config['password'],
            database=config['database'],
            docker_bin=config.get('docker_bin', 'docker'),
            container=config.get('container', ''),
            name=name
        )

        definition._init_cmds(config)
        return definition




from .service.configurationfactory import ConfigurationFactory
from .controller.backup import BackupController
from .controller.restore import RestoreController
from .controller.list import ListController
from .controller.recover import RecoverFromDisasterController
from .controller.snapshot import SnapshotController
from .mapping.handlers import HandlersMapping
from .service.client import FileRepositoryClient
from .exceptions import ApplicationException
from .service.logger import PasswordsProtectedFilter
from .service.notifier import NotifierInterface
from logging import Logger
import sys
import json


class Bahub:
    _factory = None   # type: ConfigurationFactory
    _options = {}     # type: dict
    _logger = None    # type: Logger
    _handlers = None  # type: HandlersMapping
    notifier: NotifierInterface

    def __init__(self, factory: ConfigurationFactory, options: dict, uncensored: bool, logger: Logger,
                 notifier: NotifierInterface):
        self._factory = factory
        self._options = options
        self._logger = logger
        self._handlers = HandlersMapping()
        self.notifier = notifier
        self.init_logger(uncensored)

    def run_controller(self, action_name: str, param: str, debug: bool, params: list):
        self._logger.info('Performing ' + action_name)

        try:
            if action_name == 'backup':
                controller = BackupController(
                    self._factory,
                    self._logger,
                    self._handlers,
                    FileRepositoryClient(_logger=self._logger),
                    self.notifier
                )

                print(controller.perform(param))

            elif action_name == 'restore':
                controller = RestoreController(
                    self._factory,
                    self._logger,
                    self._handlers,
                    FileRepositoryClient(_logger=self._logger),
                    self.notifier
                )

                print(controller.perform(
                        param,
                        params[2] if len(params) >= 3 else 'latest'
                    )
                )

            elif action_name in ['ls', 'list']:
                controller = ListController(
                    self._factory,
                    self._logger,
                    self._handlers,
                    FileRepositoryClient(_logger=self._logger),
                    self.notifier
                )

                response = controller.do_ls(self._factory.get_definition(param))

                print(
                    json.dumps(response, indent=4, sort_keys=True)
                )

            elif action_name == 'recover':
                controller = RecoverFromDisasterController(
                    self._factory,
                    self._logger,
                    self._handlers,
                    FileRepositoryClient(_logger=self._logger),
                    self.notifier
                )

                response = controller.perform(param)

                print(
                    json.dumps(response, indent=4, sort_keys=True)
                )

            elif action_name == 'snapshot':
                controller = SnapshotController(
                    self._factory,
                    self._logger,
                    self._handlers,
                    FileRepositoryClient(_logger=self._logger),
                    self.notifier
                )

                response = controller.perform(param)

                print(
                    json.dumps(response, indent=4, sort_keys=True)
                )

        except ApplicationException as e:
            if debug:
                raise e

            print(e)
            sys.exit(1)

    def init_logger(self, uncensored: bool):
        if not uncensored:
            self._logger.addFilter(PasswordsProtectedFilter(self._factory.get_all_sensitive_data()))
            self.notifier.set_sensitive_data_to_strip_out(self._factory.get_all_sensitive_data())





class ApplicationException(Exception):
    pass


class ConfigurationFactoryException(ApplicationException):
    pass


class ConfigurationError(ApplicationException):
    pass


class InvalidRequestException(ApplicationException):
    _json = None
    _exit_code = 0

    def __init__(self, msg, _json: dict, _exit_code: int):
        super().__init__(msg)
        self._json = _json
        self._exit_code = _exit_code

    def get_json(self) -> dict:
        return self._json


class UnexpectedResponseException(ApplicationException):
    pass


class ReadWriteException(ApplicationException):
    pass
